@@ -231,7 +231,7 @@ class ArcGISRESTfulBase:
 
         self.n_features = len(oid_list)
         if not self.disable_retry and self.verbose:
-            logger.info(f"Found {self.n_features:,} features in the requested region.")
+            logger.info(f"Found {self.n_features:,} feature(s) in the requested region.")
         return tlz.partition_all(self.max_nrecords, [str(i) for i in oid_list])  # type: ignore
 
     def _cleanup_resp(
@@ -240,10 +240,10 @@ class ArcGISRESTfulBase:
         """Remove failed responses."""
         fails = [i for i, r in enumerate(resp) if "error" in r]
 
-        if len(fails) > 0:
+        if fails:
             err = resp[fails[0]]["error"]["message"]
             resp = [r for i, r in enumerate(resp) if i not in fails]
-            if len(resp) == 0 and self.disable_retry:
+            if not resp and self.disable_retry:
                 raise ServiceError(err)
 
             if "objectIds" in payloads[fails[0]]:
@@ -251,29 +251,35 @@ class ArcGISRESTfulBase:
                 self.n_missing = len(oids)
 
                 self.failed_path = Path("cache", f"failed_ids_{self.request_id}.txt")
-                with open(self.failed_path, "w") as f:
+                with self.failed_path.open("w") as f:
                     f.write("\n".join(oids))
 
                 if not self.disable_retry:
                     self.return_m = bool(payloads[0]["ReturnM"])
                     self.return_geom = bool(payloads[0]["returnGeometry"])
                     self.total_n_features = self.n_features
-                    logger.warning(f"Found {len(fails)} failed requests. Retrying ...")
+                    if len(fails) > 1:
+                        logger.warning(f"Found {len(fails)} failed requests. Retrying ...")
+                    else:
+                        logger.warning("Found 1 failed request. Retrying ...")
                     resp.extend(self.retry_failed_requests())
 
-                    logger.warning(
-                        " ".join(
-                            [
-                                f"Total of {self.n_missing} out of {self.total_n_features}",
-                                "requested features are not available in the dataset.",
-                                "Returning the successfully retrieved features.",
-                                "The failed object IDs have been saved in the",
-                                f"file {self.failed_path}. The service returned the",
-                                "following error message for the failed requests:\n",
-                                err,
-                            ]
+                    if self.n_missing > 0:
+                        logger.warning(
+                            " ".join(
+                                [
+                                    f"Total of {self.n_missing} out of {self.total_n_features}",
+                                    "requested features are not available in the dataset.",
+                                    "Returning the successfully retrieved features.",
+                                    "The failed object IDs have been saved in the",
+                                    f"file {self.failed_path}. The service returned the",
+                                    "following error message for the failed requests:\n",
+                                    err,
+                                ]
+                            )
                         )
-                    )
+                    else:
+                        logger.info("All feature IDs have been successfully retrieved.")
 
         return resp
 
@@ -317,7 +323,7 @@ class ArcGISRESTfulBase:
         resp = self.get_response(self.query_url, payloads, "POST")
 
         resp = self._cleanup_resp(resp, payloads)
-        if len(resp) == 0:
+        if not resp:
             raise ZeroMatchedError
         return resp
 
@@ -377,9 +383,11 @@ class ArcGISRESTfulBase:
             except ServiceError:
                 continue
 
+        features = list(tlz.concat(features))
+        self.n_missing -= len(features)
         self.disable_retry = retry
         os.environ["HYRIVER_CACHE_DISABLE"] = f"{caching}".lower()
-        return list(tlz.concat(features))
+        return features  # type: ignore[return-value]
 
     def get_response(
         self, url: str, payloads: Sequence[dict[str, str]], method: str = "GET"
@@ -480,7 +488,11 @@ class WMSBase:
             raise InputValueError("outformat", self.available_outformat)
 
         if any(self.crs_str.lower() not in self.available_crs[lyr] for lyr in layers):
-            _valid_crss = (f"{lyr}: {', '.join(cs)}\n" for lyr, cs in self.available_crs.items())
+            _valid_crss = (
+                f"{lyr}: {', '.join(cs)}\n"
+                for lyr, cs in self.available_crs.items()
+                if lyr in layers
+            )
             raise InputValueError("CRS", _valid_crss)
 
     def get_validlayers(self) -> dict[str, str]:
@@ -551,7 +563,13 @@ class WFSBase:
     def __post_init__(self) -> None:
         """Validate crs."""
         self.crs_str = utils.validate_crs(self.crs)
-        self.version = validate_version(self.version, ["1.0.0", "1.1.0", "2.0.0"])
+        self.version = validate_version(self.version, ["1.1.0", "2.0.0"])
+        if self.version == "2.0.0":
+            self.nfeat_key = "numberMatched="
+            self.count_key = "count"
+        else:
+            self.nfeat_key = "numberOfFeatures="
+            self.count_key = "maxFeatures"
         valid_methods = ["json", "binary", "text"]
         if self.read_method not in valid_methods:
             raise InputValueError("read_method", valid_methods)
@@ -569,13 +587,15 @@ class WFSBase:
         self.available_layer = list(wfs.contents)
 
         wfs_features = wfs.getOperationByName("GetFeature")
-        if self.version == "1.0.0":
-            self.available_outformat = [f.rsplit("}", 1)[-1] for f in wfs_features.formatOptions]
-        else:
-            self.available_outformat = wfs_features.parameters["outputFormat"]["values"]
+        self.available_outformat = wfs_features.parameters["outputFormat"]["values"]
         self.available_outformat = [f.lower() for f in self.available_outformat]
 
-        self.available_crs = [f"{s.authority.lower()}:{s.code}" for s in wfs[self.layer].crsOptions]
+        self.available_crs = {
+            lyr: [f"{s.authority.lower()}:{s.code}" for s in wfs[lyr].crsOptions]
+            for lyr in self.available_layer
+        }
+
+        self.schema = {lyr: wfs.get_schema(lyr) for lyr in self.available_layer}
 
     def validate_wfs(self) -> None:
         """Validate input arguments with the WFS service."""
@@ -591,34 +611,34 @@ class WFSBase:
         if self.outformat not in self.available_outformat:
             raise InputValueError("outformat", self.available_outformat)
 
-        if self.crs_str.lower() not in self.available_crs:
-            raise InputValueError("crs", self.available_crs)
+        if self.crs_str.lower() not in self.available_crs[self.layer]:
+            raise InputValueError("crs", self.available_crs[self.layer])
 
-    def get_validnames(self) -> list[str]:
-        """Get valid column names for a layer."""
-        max_features = "count" if self.version == "2.0.0" else "maxFeatures"
+    def sort_params(
+        self, sort_attr: str | None, nfeatures: int, start_index: int
+    ) -> dict[str, str]:
+        """Get the sort parameters for the WFS request."""
+        if nfeatures <= self.max_nrecords:
+            return {}
 
-        payload = {
-            "service": "wfs",
-            "version": self.version,
-            "outputFormat": "application/json",
-            "request": "GetFeature",
-            "typeName": self.layer,
-            max_features: 1,
-        }
-        rjson = ar.retrieve_json([self.url], [{"params": payload}])[0]
-        valid_fields = list(
-            set(
-                utils.traverse_json(rjson, ["fields", "name"])
-                + utils.traverse_json(rjson, ["fields", "alias"])
-                + ["*"]
+        valid_attrs = self.schema[self.layer]["properties"]
+        if sort_attr is None:
+            sort_attr = next(
+                (a for a in self.schema[self.layer]["properties"] if "id" in a.lower()), None
             )
-        )
+            if sort_attr is None:
+                msg = "sort_attr is None and no id column found in the schema."
+                msg += " Please set the sort_attr manually. Available columns are:\n"
+                msg += ", ".join(list(valid_attrs))
+                raise MissingInputError(msg)
 
-        if None in valid_fields:
-            valid_fields = list(utils.traverse_json(rjson, ["features", "properties"])[0].keys())
-
-        return valid_fields
+        if sort_attr not in valid_attrs:
+            raise InputValueError("sort_attr", list(valid_attrs))
+        return {
+            "startIndex": str(start_index),
+            self.count_key: str(self.max_nrecords),
+            "sortBy": sort_attr,
+        }
 
     def __repr__(self) -> str:
         """Print the services properties."""
