@@ -6,37 +6,48 @@ import math
 import os
 from dataclasses import dataclass
 from pathlib import Path
-from typing import Any, List, Mapping, Tuple, TypeVar, Union
+from typing import Any, Generator, Mapping, TypeVar, Union
+import warnings
 
 import async_retriever as ar
+import cytoolz as tlz
 import defusedxml.ElementTree as ETree
+import joblib
 import pyproj
 import requests
-import shapely.geometry as sgeom
 import ujson as json
+import urllib3
 from requests.adapters import HTTPAdapter
 from requests.exceptions import RequestException
 from requests_cache import CachedSession, Response
 from requests_cache.backends.sqlite import SQLiteCache
 from shapely import ops
-from urllib3 import Retry
+from shapely.geometry import LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon
+from shapely.geometry import box as shapely_box
+from urllib3.exceptions import InsecureRequestWarning
 
-from .exceptions import InputTypeError, ServiceError
+from . import cache_keys
+from .exceptions import InputTypeError, InputValueError, ServiceError
 
 CRSTYPE = Union[int, str, pyproj.CRS]
 BOX_ORD = "(west, south, east, north)"
 G = TypeVar(
     "G",
-    sgeom.Point,
-    sgeom.MultiPoint,
-    sgeom.Polygon,
-    sgeom.MultiPolygon,
-    sgeom.LineString,
-    sgeom.MultiLineString,
-    Tuple[float, float, float, float],
-    List[Tuple[float, float]],
+    Point,
+    MultiPoint,
+    Polygon,
+    MultiPolygon,
+    LineString,
+    MultiLineString,
+    tuple[float, float, float, float],
+    list[tuple[float, float]],
 )
+MAX_CONN = 10
+CHUNK_SIZE = int(100 * 1024 * 1024)  # 100 MB
 
+__all__ = ["RetrySession", "traverse_json", "streaming_download", "match_crs", "validate_crs"]
+
+warnings.filterwarnings("ignore", message=".*too short worker timeout.*")
 
 def check_response(resp: str) -> str:
     """Extract error message from a response, if any."""
@@ -76,8 +87,9 @@ class RetrySession:
     expire_after : int, optional
         Expiration time for the cache in seconds, defaults to -1 (never expire).
     disable : bool, optional
-        If ``True`` temporarily disable caching requests and get new responses
-        from the server, defaults to ``False``.
+        If ``True`` temporarily disable caching request/responses, defaults to ``False``.
+    ssl : bool, optional
+        If ``True`` verify SSL certificates, defaults to ``True``.
     """
 
     def __init__(
@@ -89,21 +101,26 @@ class RetrySession:
         cache_name: str | Path | None = None,
         expire_after: int = -1,
         disable: bool = False,
+        ssl: bool = True,
     ) -> None:
-        disable = os.getenv("HYRIVER_CACHE_DISABLE", f"{disable}").lower() == "true"
-        if disable:
+        self.disable = os.getenv("HYRIVER_CACHE_DISABLE", f"{disable}").lower() == "true"
+        if self.disable:
             self.session = requests.Session()
         else:
-            self.cache_name = (
-                Path("cache", "http_cache.sqlite") if cache_name is None else Path(cache_name)
+            self.cache_name = os.getenv(
+                "HYRIVER_CACHE_NAME_HTTP", cache_name or Path("cache", "http_cache.sqlite")
             )
             backend = SQLiteCache(self.cache_name, fast_save=True, timeout=1)
             self.session = CachedSession(
                 expire_after=int(os.getenv("HYRIVER_CACHE_EXPIRE", expire_after)), backend=backend
             )
 
+        if not ssl:
+            urllib3.disable_warnings(InsecureRequestWarning)
+            self.session.verify = False
+
         adapter = HTTPAdapter(
-            max_retries=Retry(
+            max_retries=urllib3.Retry(
                 total=retries,
                 read=retries,
                 connect=retries,
@@ -119,10 +136,21 @@ class RetrySession:
         self,
         url: str,
         payload: Mapping[str, Any] | None = None,
+        params: Mapping[str, Any] | None = None,
         headers: Mapping[str, Any] | None = None,
+        stream: bool | None = None,
     ) -> Response:
         """Retrieve data from a url by GET and return the Response."""
-        resp = self.session.get(url, params=payload, headers=headers)
+        if stream and not self.disable:
+            msg = ". ".join(
+                (
+                    "Streaming is not supported with caching enabled",
+                    "Reinstantiate the class like so: RetrySession(disable=True).",
+                )
+            )
+            raise ValueError(msg)
+        params = params or payload
+        resp = self.session.get(url, params=params, headers=headers, stream=stream)
         try:
             resp.raise_for_status()
         except RequestException as ex:
@@ -134,10 +162,22 @@ class RetrySession:
         self,
         url: str,
         payload: Mapping[str, Any] | None = None,
+        data: Mapping[str, Any] | None = None,
+        json: Mapping[str, Any] | None = None,
         headers: Mapping[str, Any] | None = None,
+        stream: bool | None = None,
     ) -> Response:
         """Retrieve data from a url by POST and return the Response."""
-        resp = self.session.post(url, data=payload, headers=headers)
+        if stream and not self.disable:
+            msg = ". ".join(
+                (
+                    "Streaming is not supported with caching enabled",
+                    "Reinstantiate the class like so: RetrySession(disable=True).",
+                )
+            )
+            raise ValueError(msg)
+        data = data or payload
+        resp = self.session.post(url, data=data, json=json, headers=headers, stream=stream)
         try:
             resp.raise_for_status()
         except RequestException as ex:
@@ -148,12 +188,13 @@ class RetrySession:
     def head(
         self,
         url: str,
-        data: Mapping[str, Any] | None = None,
         params: Mapping[str, Any] | None = None,
+        data: Mapping[str, Any] | None = None,
+        json: Mapping[str, Any] | None = None,
         headers: Mapping[str, Any] | None = None,
     ) -> Response:
         """Retrieve data from a url by POST and return the Response."""
-        resp = self.session.head(url, data=data, params=params, headers=headers)
+        resp = self.session.head(url, data=data, params=params, json=json, headers=headers)
         try:
             resp.raise_for_status()
         except RequestException as ex:
@@ -161,6 +202,140 @@ class RetrySession:
         else:
             return resp
 
+    def close(self) -> None:
+        """Close the session."""
+        self.session.close()
+
+
+def _prepare_requests_args(
+    urls: list[str] | str,
+    kwds: list[dict[str, dict[Any, Any]]] | dict[str, dict[Any, Any]] | None,
+    method: str,
+    fnames: str | Path | list[str | Path] | None,
+    file_prefix: str,
+    file_extention: str,
+) -> tuple[
+    tuple[str, ...], tuple[dict[str, None | dict[Any, Any]], ...], Generator[Path, None, None]
+]:
+    """Get url and kwds for streaming download."""
+    url_list = tuple(urls) if isinstance(urls, (list, tuple)) else (urls,)
+
+    if kwds is None:
+        if method == "GET":
+            kwd_list = ({"params": None},) * len(url_list)
+        else:
+            kwd_list = ({"data": None},) * len(url_list)
+    else:
+        kwd_list = tuple(kwds) if isinstance(kwds, (list, tuple)) else (kwds,)  # type: ignore
+    key_list = {k for keys in kwd_list for k in keys}
+    valid_keys = ("params", "data", "json", "headers")
+    if any(k not in valid_keys for k in key_list):
+        raise InputValueError("kwds", valid_keys)
+
+    if len(url_list) != len(kwd_list):
+        raise InputTypeError("urls/kwds", "list of same length")
+
+    fex = file_extention.replace(".", "")
+    if fnames is None:
+        cache_dir = os.getenv("HYRIVER_CACHE_NAME", Path("cache", "tmp"))
+        cache_dir = Path(cache_dir).parent
+        files = (
+            Path(cache_dir, f"{file_prefix}{cache_keys.create_key(method, u, **p)}.{fex}")
+            for u, p in zip(url_list, kwd_list)
+        )
+    else:
+        f_list = tuple(fnames) if isinstance(fnames, (list, tuple)) else (fnames,)
+        if len(url_list) != len(f_list):
+            raise InputTypeError("urls/fnames", "lists of same length")
+        files = (Path(f) for f in f_list)
+
+    return url_list, kwd_list, files
+
+
+def streaming_download(
+    urls: list[str] | str,
+    kwds: list[dict[str, dict[Any, Any]]] | dict[str, dict[Any, Any]] | None = None,
+    fnames: str | Path | list[str | Path] | None = None,
+    file_prefix: str = "",
+    file_extention: str = "",
+    method: str = "GET",
+    ssl: bool = True,
+    chunk_size: int = CHUNK_SIZE,
+    n_jobs: int = MAX_CONN,
+) -> Path | list[Path]:
+    """Download and store files in parallel from a list of URLs/Keywords.
+
+    Notes
+    -----
+    This function uses ``joblib`` with ``loky`` backend.
+
+    Parameters
+    ----------
+    urls : tuple or list
+        A list of URLs to download.
+    kwds : tuple or list, optional
+        A list of keywords associated with each URL, e.g.,
+        ({"params": ..., "headers": ...}, ...). Defaults to ``None``.
+    fnames : tuple or list, optional
+        A list of filenames associated with each URL, e.g.,
+        ("file1.zip", ...). Defaults to ``None``. If not provided,
+        random unique filenames will be generated based on
+        URL and keyword pairs.
+    file_prefix : str, optional
+        Prefix to add to filenames when storing the files, defaults
+        to ``None``, i.e., no prefix. This argument will be only be
+        used if ``fnames`` is not passed.
+    file_extention : str, optional
+        Extension to use for storing the files, defaults to ``None``,
+        i.e., no extension if ``fnames`` is not provided otherwise. This
+        argument will be only be used if ``fnames`` is not passed.
+    method : str, optional
+        HTTP method to use, i.e, ``GET`` or ``POST``, by default "GET".
+    ssl : bool, optional
+        Whether to use SSL verification, by default ``True``.
+    chunk_size : int, optional
+        Chunk size to use when downloading, by default 100 MB is used.
+    n_jobs: int, optional
+        The maximum number of concurrent downloads, defaults to 10.
+
+    Returns
+    -------
+    list
+        A list of ``pathlib.Path`` objects associated with URLs in the
+        same order.
+    """
+    method = method.upper()
+    valid_methods = ("GET", "POST")
+    if method not in valid_methods:
+        raise InputValueError("method", valid_methods)
+
+    url_list, kwd_list, files = _prepare_requests_args(
+        urls, kwds, method, fnames, file_prefix, file_extention
+    )
+
+    session = RetrySession(disable=True, ssl=ssl)
+    if method == "GET":
+        func = tlz.partial(session.get, stream=True)
+    else:
+        func = tlz.partial(session.post, stream=True)
+
+    def download(url: str, kwd: Mapping[str, None | dict[Any, Any]], fname: Path) -> Path:
+        resp = func(url, **kwd)
+        fsize = int(resp.headers.get("Content-Length", -1))
+        if not fname.exists() or fname.stat().st_size != fsize:
+            fname.parent.mkdir(exist_ok=True, parents=True)
+            with fname.open("wb") as f:
+                f.writelines(resp.iter_content(chunk_size))
+        return fname
+
+    if not isinstance(urls, (list, tuple)):
+        return download(url_list[0], kwd_list[0], next(files))
+
+    fpaths: list[Path] = joblib.Parallel(n_jobs=n_jobs)(
+        joblib.delayed(download)(u, k, f) for u, k, f in zip(url_list, kwd_list, files)
+    )
+    return fpaths
+
 
 def traverse_json(
     items: dict[str, Any] | list[dict[str, Any]], ipath: str | list[str]
@@ -231,9 +406,7 @@ def traverse_json(
     if isinstance(items, dict):
         return extract(items, ipath, 0, [])
 
-    outer_arr = []
-    for item in items:
-        outer_arr.append(extract(item, ipath, 0, []))
+    outer_arr = [extract(item, ipath, 0, []) for item in items]
     return outer_arr
 
 
@@ -243,9 +416,9 @@ class ESRIGeomQuery:
 
     Parameters
     ----------
-    geometry : tuple or sgeom.Polygon or sgeom.Point or sgeom.LineString
+    geometry : tuple or Polygon or Point or LineString
         The input geometry which can be a point (x, y), a list of points [(x, y), ...],
-        bbox (xmin, ymin, xmax, ymax), or a Shapely's sgeom.Polygon.
+        bbox (xmin, ymin, xmax, ymax), or a Shapely's Polygon.
     wkid : int
         The Well-known ID (WKID) of the geometry's spatial reference e.g., for EPSG:4326,
         4326 should be passed. Check
@@ -257,8 +430,8 @@ class ESRIGeomQuery:
         tuple[float, float]
         | list[tuple[float, float]]
         | tuple[float, float, float, float]
-        | sgeom.Polygon
-        | sgeom.LineString
+        | Polygon
+        | LineString
     )
     wkid: int
 
@@ -313,7 +486,7 @@ class ESRIGeomQuery:
 
     def polygon(self) -> Mapping[str, str]:
         """Query for a polygon."""
-        if not isinstance(self.geometry, sgeom.Polygon):
+        if not isinstance(self.geometry, Polygon):
             raise InputTypeError("geometry", "Polygon")
 
         geo_type = "esriGeometryPolygon"
@@ -322,7 +495,7 @@ class ESRIGeomQuery:
 
     def polyline(self) -> Mapping[str, str]:
         """Query for a polyline."""
-        if not isinstance(self.geometry, sgeom.LineString):
+        if not isinstance(self.geometry, LineString):
             raise InputTypeError("geometry", "LineString")
 
         geo_type = "esriGeometryPolyline"
@@ -371,18 +544,18 @@ def match_crs(geom: G, in_crs: CRSTYPE, out_crs: CRSTYPE) -> G:
     if isinstance(
         geom,
         (
-            sgeom.Polygon,
-            sgeom.LineString,
-            sgeom.MultiLineString,
-            sgeom.MultiPolygon,
-            sgeom.Point,
-            sgeom.MultiPoint,
+            Polygon,
+            LineString,
+            MultiLineString,
+            MultiPolygon,
+            Point,
+            MultiPoint,
         ),
     ):
         return ops.transform(project, geom)  # type: ignore
 
     if isinstance(geom, tuple) and len(geom) == 4:
-        return ops.transform(project, sgeom.box(*geom)).bounds  # type: ignore
+        return ops.transform(project, shapely_box(*geom)).bounds  # type: ignore
 
     if isinstance(geom, list) and all(len(c) == 2 for c in geom):
         xx, yy = zip(*geom)
@@ -452,8 +625,8 @@ def bbox_decompose(
 
     xmin, ymin, xmax, ymax = match_crs(bbox, box_crs, 4326)
 
-    x_dist = geod.geometry_length(sgeom.LineString([(xmin, ymin), (xmax, ymin)]))
-    y_dist = geod.geometry_length(sgeom.LineString([(xmin, ymin), (xmin, ymax)]))
+    x_dist = geod.geometry_length(LineString([(xmin, ymin), (xmax, ymin)]))
+    y_dist = geod.geometry_length(LineString([(xmin, ymin), (xmin, ymax)]))
     width = int(math.ceil(x_dist / resolution))
     height = int(math.ceil(y_dist / resolution))
 
